/*****************************************************************************
 *
 * @defgroup
 *
 * description:
 *  This phase will receive a fold which should look like the following:
 *  
 *      a = with {
 *          ( lb <= iv < ub )
 *          {
 *              acc = _accu_ (iv, neutr);
 *              val = expr (iv);
 *              v1, v2 = _accu_dummy_ (iv, neutr);
 *              cpu_res = foldop(acc, val);
 *              gpu_res = foldop(v1, v2);
 *              res = _accu_sel_(cpu_res, gpu_res, val);
 *          } : res;
 *      } : fold (foldop (), neutr);
 *      
 *      CPU CODE:
 *  If the fold cannot be cudarized, we will revert it back to it's original
 *  state by changing the code block expression to the first argument of the
 *  associated _accu_sel_. After this, the regular CPU code generation will 
 *  take care of the rest. After deadcoderemoval this should result in the code
 *  below:
 * 
 *      a = with {
 *          ( lb <= iv < ub )
 *          {
 *              acc = _accu_ (iv, neutr);
 *              val = expr (iv);
 *              res' = foldop(acc, val);
 *          } : res';
 *      } : fold (foldop (), neutr);
 * 
 *      GPU CODE:
 *  If it was decided we will run this on the device, we will drastically 
 *  change the with-loop. On a global level, four major changes will happen:
 *  
 *  1)  The original foldop is lifted into a separate function. This is done as
 *      during optimisation the foldop could have been split into any arbitrary 
 *      number of assignments. To ensure that the foldop can be accessed in a 
 *      single place, we lift the optimised version to a separate function. The
 *      code below shows how such a lifted function will roughly look:
 *      
 *      TYPE(fres) CUDA_Fold_foldop (TYPE(V1) v1, TYPE(V2) v2) {
 *          return foldop(v1, v2);
 *      }
 *      
 *  2)  Before the with-loop, the expected CUDA grid size is calculated. This
 *      gridsize is then used to generate a temporary foldbuffer (fb), which
 *      stores the partial results from the partial folding.
 *  3)  A second GPU fold with-loop is appended after the primary fold 
 *      with-loop referred to as the 'final with-loop'. This fold will fold the
 *      partial results from the first fold into a single value. It is able to 
 *      do this as it will only have a single thread block, hence one partial 
 *      result. Along with this final fold, a foldbuffer of size 1 is allocated
 *      to store the 'partial' result of this fold.
 *  4)  Finally, the result from the final fold is written to the original avis 
 *      of the original fold with-loop.
 *  
 *  The bodies of both with-loop bodies are very similar, we will describe how
 *  the first fold looks after this phase and then how the final fold differs.
 *  
 *  After the CUFD phase and deadcoderemoval the first fold should look like  
 *  the following:
 *      
 *      foldbuf = with {
 *          ( lb <= iv < ub )
 *          {
 *              sdata = _init_shared_(int_shmem[1024], neutr);
 *              bx = _cuda_blockIdx_x_();
 *              tx = _cuda_threadIdx_x_();
 *              cond = _eq_SxS_ (tx, 0);
 *              val = expr (iv);
 *              sdata = _idx_modarray_A_S_S_ (sdata, tx, val);
 *              sdata = _cuda_fold_block_ (tx, sdata);
 *              res = _cond_wl_assign_ (cond, tx, sdata, bx, foldbuf);
 *          } : res;
 *      } : fold (CUDA_Fold_foldop(), neutr);
 *  
 *  _cuda_fold_block_ has the PRF_FOLDOP attribute set to the CUDA_Fold_foldop.
 *  The _cond_wl_assign_ will assign the value from sdata[tx] to foldbuf[bx] if
 *  the cond variable holds (in this case, if tx == 0).
 *  
 *  The final fold differs in some manners, and looks the following:
 *  
 *      foldbuf' = with {
 *          ( [0] <= iv=[i] < [_min_SxS_(grid, 1024)] )
 *          {
 *              sdata = _init_shared_(int_shmem[1024], neutr);
 *              val = _idx_sel_ (i, foldbuf);
 *              offset = _add_SxS_ (i, 1024);
 *              cond = _eq_SxS_ (i, 0);
 *              sdata = _idx_modarray_A_S_S_ (sdata, i, val);
 *              sdata = _cuda_prefold_linear_ (i, offset, grid, foldbuf, sdata);
 *              sdata = _cuda_fold_block_ (i, sdata);
 *              res = _cond_wl_assign_ (cond, i, sdata, 0, foldbuf');
 *          } : res;
 *      } : fold (CUDA_Fold_foldop(), neutr);
 *  
 *  As the partition range is from 0 to 1024, this with-loop will only generate
 *  a single kernel. Here we generate either grid or 1024 kernels. This is done
 *  as when gird is less than 1024, we would go out of bounds when selecting
 *  val from the foldbuf. As it is only a single kernel, we can use i as a 
 *  substitute of tx. 
 *  As we are folding the previous partial results, we can simply select values
 *  from the previous foldbuffer. If the gridsize is greater than 1024, we use
 *  _cuda_prefold_linear_ to prefold all values greater than 1024 back into the
 *  sdata (through a simple for loop for every index).
 *  
 *  Finally, we have a simple selection of the only value left and store it in
 *  the result variable:
 *  
 *      a = _idx_sel_(0, foldbuf');
 *  
 *  With-loops with multiple folds are handled almost the same way, rather than
 *  go into full detail, we will simply show the transformation. Initially we 
 *  start with the following folds:
 *
 *      a = with {
 *          ( lb <= iv < ub )
 *          {
 *              acc1, acc2 = _accu_ (iv, neutr1, neutr2);
 *              val1 = expr1 (iv);
 *              val2 = expr2 (iv);
 *              v1a, v2a, v1b, v2b = _accu_dummy_ (iv, neutr1, neutr2);
 *              cpu_res1 = foldop1(acc1, val1);
 *              cpu_res2 = foldop2(acc2, val2);
 *              gpu_res1 = foldop1(v1a, v2a);
 *              gpu_res2 = foldop2(v1b, v2b);
 *              res1 = _accu_sel_(cpu_res1, gpu_res1, val1);
 *              res2 = _accu_sel_(cpu_res2, gpu_res2, val2);
 *          } : res1, res2;
 *      } : fold (foldop1(), neutr1),
 *          fold (foldop2(), neutr2);
 *  
 *  The code after this phase should be as follows:
 *  
 *      foldbuf1, foldbuf2 = with {
 *          ( lb <= iv < ub )
 *          {
 *              sdata1 = _init_shared_(int_shmem[1024], neutr1);
 *              sdata2 = _init_shared_(int_shmem[1024], neutr2);
 *              bx = _cuda_blockIdx_x_();
 *              tx = _cuda_threadIdx_x_();
 *              val1 = expr1 (iv);
 *              val2 = expr2 (iv);
 *              cond = _eq_SxS_ (tx, 0);
 *              sdata1 = _idx_modarray_A_S_S_ (sdata1, tx, val1);
 *              sdata2 = _idx_modarray_A_S_S_ (sdata2, tx, val2);
 *              sdata1 = _cuda_fold_block_ (tx, sdata1);
 *              sdata2 = _cuda_fold_block_ (tx, sdata2);
 *              res1 = _cond_wl_assign_ (cond, tx, sdata1, bx, foldbuf2);
 *              res2 = _cond_wl_assign_ (cond, tx, sdata2, bx, foldbuf2);
 *          } : res1, res2;
 *      } : fold (CUDA_Fold_foldop1(), neutr1),
 *          fold (CUDA_Fold_foldop2(), neutr2);
 *  
 *      foldbuf1', foldbuf2' = with {
 *          ( [0] <= iv=[i] < [_min_SxS_(grid, 1024)] )
 *          {
 *              sdata1 = _init_shared_(int_shmem[1024], neutr1);
 *              sdata2 = _init_shared_(int_shmem[1024], neutr2);
 *              val1 = _idx_sel_ (i, foldbuf1);
 *              val2 = _idx_sel_ (i, foldbuf2);
 *              offset = _add_SxS_ (i, 1024);
 *              cond = _eq_SxS_ (i, 0);
 *              sdata1 = _idx_modarray_A_S_S_ (sdata1, i, val1);
 *              sdata2 = _idx_modarray_A_S_S_ (sdata2, i, val2);
 *              sdata1 = _cuda_prefold_linear_ (i, offset, grid, foldbuf1, sdata1);
 *              sdata2 = _cuda_prefold_linear_ (i, offset, grid, foldbuf2, sdata2);
 *              sdata1 = _cuda_fold_block_ (i, sdata1);
 *              sdata2 = _cuda_fold_block_ (i, sdata2);
 *              res1 = _cond_wl_assign_ (cond, tx, sdata1, bx, foldbuf2);
 *              res2 = _cond_wl_assign_ (cond, tx, sdata2, bx, foldbuf2);
 *          } : res1, res2;
 *      } : fold (CUDA_Fold_foldop1(), neutr1),
 *          fold (CUDA_Fold_foldop2(), neutr2);
 *      a = _idx_sel_( 0, foldbuf1');
 *      b = _idx_sel_( 0, foldbuf2');
 *  
 *  Side note I didn't write into text: Partial folds always get the pragma 
 *    #pragma gpukernel GridBlock(1, SplitLast(1024, CompressGrid([1], ShiftLB(Gen))))
 * 
 *****************************************************************************/
 
/** <!--********************************************************************-->
 *
 * @file cuda_fold.c
 *
 * Prefix: CUFD
 *
 *****************************************************************************/
 
#include "cuda_fold.h"

#include <stdlib.h>
#include "tree_basic.h"
#include "tree_compound.h"
#include "str.h"
#include "memory.h"
#include "globals.h"

#define DBUG_PREFIX "CUFD"
#include "debug.h"

#include "ctinfo.h"
#include "traverse.h"
#include "free.h"
#include "namespaces.h"
#include "new_types.h"
#include "DupTree.h"
#include "type_utils.h"
#include "shape.h"
#include "cuda_utils.h"

#include "types.h"
#include "vector.h"
#include "constants.h"

#include "LookUpTable.h"

#include "deserialize.h"
#include "new_typecheck.h"
#include "deadcoderemoval.h"
#include "unused_argument_annotate.h"


// Remove me later
#include "print.h"

/** <!--********************************************************************-->
 *
 * @name INFO structure
 * @{
 *
 *****************************************************************************/
struct INFO {
    node *fundef;
    node *assign;
    node *let_lhs;
    node *with;
    bool first_fold;
    
    node *sdata_assigns;
    node *cuda_fold_assigns;
    node *res_assigns;
    node *grid_avis;
    node *tx_avis;
    node *bx_avis;
    node *cond_avis;
    
    node *final_with;
    node *final_with_assign;
    node *final_index_avis;
    node *final_cond_avis;
    node *final_prefold_start_avis;
    node *final_sdata_assigns;
    node *final_prefold_assigns;
    node *final_cuda_fold_assigns;
    node *final_res_assigns;
    
    node *cexprs;
    node *accu_dummy_ids;
    node *postwith_assigns;
    node *prewith_assigns;
    
    node *fun_vardecs;
    node *fun_args;
    node *fun_akv_ids;
    node *gpu_funs;
    
    lut_t *lut;
    lut_t *lut_inv;
};

#define INFO_FUNDEF(n) (n->fundef)              // Current fundef
#define INFO_ASSIGN(n) (n->assign)              // Current assign
#define INFO_LET_LHS(n) (n->let_lhs)            // Ids from current let
#define INFO_WITH(n) (n->with)                  // Current withloop
#define INFO_FIRST_FOLD(n) (n->first_fold)      // If this is the first fold

#define INFO_SDATA_ASSIGNS(n) (n->sdata_assigns)// Assings for sdata
#define INFO_CUDA_FOLD_ASSIGNS(n) (n->cuda_fold_assigns)
                                                // Assigns for the cuda_folds
#define INFO_RES_ASSIGNS(n) (n->res_assigns)    // The result assignments
#define INFO_GRID_AVIS(n) (n->grid_avis)        // Avis that holds the grid value
#define INFO_TX_AVIS(n) (n->tx_avis)            // Avis that hold the thread x value
#define INFO_BX_AVIS(n) (n->bx_avis)            // Avis that hold the block x value
#define INFO_COND_AVIS(n) (n->cond_avis)        // Avis that hold the tx == 0 value

#define INFO_FINAL_WITH(n) (n->final_with)      // Contains second with loop
                                                // Contains second fold with loop
#define INFO_FINAL_WITH_ASSIGN(n) (n->final_with_assign)
                                                // Contains second fold with loop
#define INFO_FINAL_INDEX_AVIS(n) (n->final_index_avis)
                                                // Avis i for withid (iv = [i])
#define INFO_FINAL_COND_AVIS(n) (n->final_cond_avis)
                                                // Avis that holds the i == 0 value
#define INFO_FINAL_OFFSET_START_AVIS(n) (n->final_prefold_start_avis)
                                                // Avis that holds i + 1024
#define INFO_FINAL_SDATA_ASSIGNS(n) (n->final_sdata_assigns)
                                                // Sdata used for second generated fold
#define INFO_FINAL_PREFOLD_ASSIGNS(n) (n->final_prefold_assigns)
                                                // Prefolds used for second generated fold
#define INFO_FINAL_CUDA_FOLD_ASSIGNS(n) (n->final_cuda_fold_assigns)
                                                // Partial folds used for second generated fold
#define INFO_FINAL_RES_ASSIGNS(n) (n->final_res_assigns)
                                                // Res assings used for second generated fold


#define INFO_CEXPRS(n) (n->cexprs)              // Current cexpr
#define INFO_ACCU_DUMMY_IDS(n) (n->accu_dummy_ids)
                                                // Current accu dummy lhs ids
#define INFO_POSTWITH_ASSINGS(n) (n->postwith_assigns)
                                                // Assings to be appended directly after with
#define INFO_PREWITH_ASSINGS(n) (n->prewith_assigns)
                                                // Assings to be prepended before with
                                                
#define INFO_FUN_VARDECS(n) (n->fun_vardecs)    // Vardecs of a new function
#define INFO_FUN_ARGS(n) (n->fun_args)          // Args of a new function
#define INFO_FUN_AKV_IDS(n) (n->fun_akv_ids)    // Args which are akv, inlined as vardecs
#define INFO_GPU_FUNS(n) (n->gpu_funs)          // New device functions

#define INFO_LUT(n) (n->lut)                    // Lut used to create the device function
#define INFO_LUT_INV(n) (n->lut)                // Inverse of lut, gets device function arguments



static info *
MakeInfo (void)
{
    info *result;

    DBUG_ENTER ();

    result = (info *)MEMmalloc (sizeof (info));

    INFO_FUNDEF (result) = NULL;
    INFO_ASSIGN (result) = NULL;
    INFO_LET_LHS (result) = NULL;
    INFO_WITH (result) = NULL;
    INFO_FIRST_FOLD (result) = TRUE;
    
    INFO_SDATA_ASSIGNS (result) = NULL;
    INFO_CUDA_FOLD_ASSIGNS (result) = NULL;
    INFO_RES_ASSIGNS (result) = NULL;
    INFO_GRID_AVIS (result) = NULL;
    INFO_TX_AVIS (result) = NULL;
    INFO_BX_AVIS (result) = NULL;
    INFO_COND_AVIS (result) = NULL;
    
    INFO_FINAL_WITH (result) = NULL;
    INFO_FINAL_WITH_ASSIGN (result) = NULL;
    INFO_FINAL_INDEX_AVIS (result) = NULL;
    INFO_FINAL_COND_AVIS (result) = NULL;
    INFO_FINAL_OFFSET_START_AVIS (result) = NULL;
    INFO_FINAL_SDATA_ASSIGNS (result) = NULL;
    INFO_FINAL_PREFOLD_ASSIGNS (result) = NULL;
    INFO_FINAL_CUDA_FOLD_ASSIGNS (result) = NULL;
    INFO_FINAL_RES_ASSIGNS (result) = NULL;
    
    INFO_CEXPRS (result) = NULL;
    INFO_ACCU_DUMMY_IDS (result) = NULL;
    INFO_POSTWITH_ASSINGS (result) = NULL;
    INFO_PREWITH_ASSINGS (result) = NULL;
    
    INFO_FUN_VARDECS (result) = NULL;
    INFO_FUN_ARGS (result) = NULL;
    INFO_FUN_AKV_IDS (result) = NULL;
    INFO_GPU_FUNS (result) = NULL;
    
    INFO_LUT (result) = NULL;
    INFO_LUT_INV (result) = NULL;
    
    DBUG_RETURN (result);
}

static info *
FreeInfo (info *info)
{
    DBUG_ENTER ();

    info = MEMfree (info);

    DBUG_RETURN (info);
}

/** <!--********************************************************************-->
 * @}  <!-- INFO structure -->
 *****************************************************************************/ 

/** <!--********************************************************************-->
 *
 * @name Entry functions
 * @{
 *
 *****************************************************************************/

node *
CUFDdoCudaFold(node *syntax_tree)
{
    info *info;
    DBUG_ENTER ();


    DBUG_ASSERT (NODE_TYPE (syntax_tree) == N_module,
                 "CUFDdoCudaFold not started with module node");
    DBUG_PRINT ("Starting CudaFold");
    
    info = MakeInfo ();
    TRAVpush (TR_cufd);
    syntax_tree = TRAVdo (syntax_tree, info);
    TRAVpop ();
    info = FreeInfo (info);
    
    DBUG_PRINT ("Finished CudaFold");
    
    DBUG_RETURN (syntax_tree);
}

/** <!--********************************************************************-->
 * @}  <!-- Entry functions -->
 *****************************************************************************/

/** <!--********************************************************************-->
 *
 * @name Static helper functions
 * @{
 *
 *****************************************************************************/
// Generic helper functions
/******************************************************************************
 *
 * function:
 *    ntype *CreateScalarType (simpletype base)
 *
 * description:
 *    Creates the scalar type based on the provided base.
 *
 ******************************************************************************/
static ntype *
CreateScalarType (simpletype base)
{
    ntype *type;
    DBUG_ENTER ();
    type = TYmakeAKS (TYmakeSimpleType (base), SHmakeShape (0));
    DBUG_RETURN (type);
}

/******************************************************************************
 *
 * function:
 *    node *GetFirstIdsAvis (node *assign_chain)
 *
 * description:
 *    Retrieves the LHS avis of the first assign.
 *
 ******************************************************************************/
static node *
GetFirstIdsAvis (node *assign_chain)
{
    node *avis;
    DBUG_ENTER ();
    avis = assign_chain == NULL ? NULL : IDS_AVIS (ASSIGN_LHS (assign_chain));
    DBUG_RETURN (avis);
}
/******************************************************************************
 *
 * function:
 *    node *GetLastIdsAvis (node *assign_chain)
 *
 * description:
 *    Retrieves the last assign and its LHS avis.
 *
 ******************************************************************************/
static node *
GetLastIdsAvis (node *assign_chain)
{
    node *assign, *avis;
    DBUG_ENTER ();
    assign = TCgetLastAssign (assign_chain);
    avis = assign == NULL ? NULL : IDS_AVIS (ASSIGN_LHS (assign));
    DBUG_RETURN (avis);
}

/******************************************************************************
 *
 * function:
 *    node *FindAssignLetPrf ( node *assigns, prf target_prf)
 *
 * description:
 *    Finds the first top level N_let assign that uses the target_prf in the
 *    RHS.
 *
 ******************************************************************************/
static node *
FindAssignLetPrf (node *assigns, prf target_prf)
{
    node *stmt;
    node *ret;
    DBUG_ENTER ();
    
    if (assigns != NULL) {
        stmt = ASSIGN_STMT (assigns);
        // Let with our prf
        if ((ASSIGN_STMTTYPE (assigns) == N_let) &&
            (NODE_TYPE (ASSIGN_RHS (assigns)) == N_prf) &&
            (PRF_PRF (ASSIGN_RHS (assigns)) == target_prf)) {
            ret = assigns;
        } else {
            ret = FindAssignLetPrf (ASSIGN_NEXT (assigns), target_prf);
        }
    } else {
        ret = NULL;
    }
    
    DBUG_RETURN (ret);
}
/******************************************************************************
 *
 * function:
 *    node *RemoveAssign (node *assigns, node *target_assign)
 *
 * description:
 *    If the target_assign is found in the assings chian, it is removed. If the
 *    assign is an N_let the LHS ids SSAASSIGN is set to NULL. 
 *
 ******************************************************************************/
static node *
RemoveAssign (node *assigns, node *target_assign)
{
    node *ids;
    DBUG_ENTER ();
    
    if (assigns != NULL) {
        if (assigns == target_assign) {
            if (ASSIGN_STMTTYPE (assigns) == N_let) {
                ids = ASSIGN_LHS (assigns);
                while (ids != NULL) {
                    AVIS_SSAASSIGN (IDS_AVIS (ids)) = NULL;
                    ids = IDS_NEXT (ids);
                }
            }
            assigns = FREEdoFreeNode (assigns);
        } else {
            ASSIGN_NEXT (assigns)
                = RemoveAssign (ASSIGN_NEXT (assigns), target_assign);
        }
    }
    
    DBUG_RETURN (assigns);
}
/******************************************************************************
 *
 * function:
 *    node *MakeAvisAndVardec (char *avis_name, ntype *avis_type, info *arg_info)
 *
 * description:
 *    Creates a basic avis and adds it the to vardecs of the fundef in arg_info.
 *
 ******************************************************************************/
static node *
MakeAvisAndVardec (char *avis_name, ntype *avis_type, info *arg_info)
{
    node *res;
    DBUG_ENTER ();
    
    res = TBmakeAvis (avis_name, avis_type);
    FUNDEF_VARDECS (INFO_FUNDEF (arg_info)) 
        = TBmakeVardec (res, FUNDEF_VARDECS (INFO_FUNDEF (arg_info)));
    
    DBUG_RETURN (res);
}
/******************************************************************************
 *
 * function:
 *    node *CopyAvisAndVardec (node *avis) 
 *
 * description:
 *    Creates a basic copy of the provided avis and adds it as a vardec. 
 *    Only type and name are set.
 *
 ******************************************************************************/
static node *
CopyAvisAndVardec (node *avis, info *arg_info) 
{
    node *avis_copy;
    DBUG_ENTER ();
    avis_copy = MakeAvisAndVardec (TRAVtmpVarName (AVIS_NAME (avis)),
                                   TYcopyType (AVIS_TYPE (avis)),
                                   arg_info);
    AVIS_SHAREDINITVALUE (avis_copy) = AVIS_SHAREDINITVALUE (avis);

    DBUG_RETURN (avis_copy);
}
/******************************************************************************
 *
 * function:
 *    node *MakeAssignLet (node *lhs_ids, node *rhs_expr, node *next_assign)
 *
 * description:
 *    Creates an let assign using the provided arguments.
 *
 ******************************************************************************/
static node *
MakeAssignLet (node *lhs_ids, node *rhs_expr, node *next_assign)
{
    node *res, *ids;
    DBUG_ENTER ();
    
    res = TBmakeAssign (TBmakeLet (lhs_ids, rhs_expr), next_assign);
    ids = lhs_ids;
    while (ids != NULL) {
        AVIS_SSAASSIGN (IDS_AVIS (ids)) = res;
        ids = IDS_NEXT (ids);
    }
    
    DBUG_RETURN (res);
}
/******************************************************************************
 *
 * function:
 *    node *AppendLetAssign (char *name, ntype *type, node *rhs, node *assign_chain, info *arg_info)
 *
 * description:
 *    Appends a new assign with an avis defined by the name and type on the LHS
 *    and the RHS argument on the RHS.
 *
 ******************************************************************************/
static node *
AppendLetAssign (char *name, ntype *type, node *rhs, node *assign_chain, info *arg_info)
{
    node *avis, *assign;
    DBUG_ENTER ();
    avis = MakeAvisAndVardec (TRAVtmpVarName (name), type, arg_info);
    assign = TBmakeAssign (TBmakeLet (TBmakeIds (avis, NULL), rhs), NULL);
    AVIS_SSAASSIGN (avis) = assign;
    assign_chain = TCappendAssign (assign_chain, assign);
    DBUG_RETURN (assign_chain);
}
/******************************************************************************
 *
 * function:
 *    node *PreppendLetAssign (char *name, ntype *type, node *rhs, node *next_assign, info *arg_info)
 *
 * description:
 *    Preppends a new assign with an avis defined by the name and type on the 
 *    LHS and the RHS argument on the RHS.
 *
 ******************************************************************************/
static node *
PreppendLetAssign (char *name, ntype *type, node *rhs, node *next_assign, info *arg_info)
{
    node *avis;
    DBUG_ENTER ();
    avis = MakeAvisAndVardec (TRAVtmpVarName (name), type, arg_info);
    next_assign = TBmakeAssign (TBmakeLet (TBmakeIds (avis, NULL), rhs), next_assign);
    AVIS_SSAASSIGN (avis) = next_assign;
    DBUG_RETURN (next_assign);
}
/******************************************************************************
 *
 * function:
 *    node *AddToLuts (node *old, node *new, info *arg_info) 
 *
 * description:
 *    Adds the relation between the old and new nodes in both the INFO_LUT and 
 *    the inverse relation in INFO_LUT_INV.
 *
 ******************************************************************************/
static node *
AddToLuts (node *old, node *new, info *arg_info) 
{
    DBUG_ENTER ();
    INFO_LUT (arg_info) = LUTinsertIntoLutP (INFO_LUT (arg_info), 
                                             old, new);
    INFO_LUT_INV (arg_info) = LUTinsertIntoLutP (INFO_LUT_INV (arg_info), 
                                                 new, old);
    DBUG_RETURN (new);
}
/******************************************************************************
 *
 * function:
 *    node *AvisCopyAsFunArgs (node *avis, info *arg_info) 
 *
 * description:
 *    Creates a copy of the provided avis, adds the avis and its copy to the
 *    LUTs and finally creates an arg from the new avis.
 *
 ******************************************************************************/
static node *
AvisCopyAsFunArgs (node *avis, info *arg_info) 
{
    node *new_avis;
    DBUG_ENTER ();
    
    new_avis = TBmakeAvis (TRAVtmpVarName (AVIS_NAME (avis)), TYcopyType (AVIS_TYPE (avis)));
    new_avis = AddToLuts (avis, new_avis, arg_info);
    INFO_FUN_ARGS (arg_info) = TBmakeArg (new_avis, INFO_FUN_ARGS(arg_info));
    
    DBUG_RETURN (new_avis);
}
/******************************************************************************
 *
 * function:
 *    node *AvisCopyAsFunVardecs (node *avis, info *arg_info) 
 *
 * description:
 *    Creates a copy of the provided avis, adds the avis and its copy to the
 *    LUTs and finally creates an vardec from the new avis.
 *
 ******************************************************************************/
static node *
AvisCopyAsFunVardecs (node *avis, info *arg_info) 
{
    node *new_avis;
    DBUG_ENTER ();
    
    new_avis = TBmakeAvis (TRAVtmpVarName (AVIS_NAME (avis)), TYcopyType (AVIS_TYPE (avis)));
    new_avis = AddToLuts (avis, new_avis, arg_info);
    INFO_FUN_VARDECS (arg_info) = TBmakeVardec (new_avis, INFO_FUN_VARDECS(arg_info));
    
    DBUG_RETURN (new_avis);
}
/******************************************************************************
 *
 * function:
 *    node *RemoveUnusedArgs (node *args)
 *
 * description:
 *    Removes any args from the provided args that aren't used in the body.
 *
 ******************************************************************************/
static node *
RemoveUnusedArgs (node *args)
{
    DBUG_ENTER ();
    if (args != NULL) {
        if(ARG_ISUSEDINBODY (args)) {
            ARG_NEXT (args) = RemoveUnusedArgs (ARG_NEXT (args));
        } else {
            args = FREEdoFreeNode (args);
            args = RemoveUnusedArgs (args);
        }
    }
    DBUG_RETURN (args);
}

// Specific helper functions
node *
ATravIdFun (node *arg_node, info *arg_info)
{
    node *avis, *new_avis;
    DBUG_ENTER ();
    
    avis = ID_AVIS (arg_node);
    new_avis = LUTsearchInLutPp (INFO_LUT (arg_info), avis);
    // Not found in lut
    if (new_avis == avis) {
        if (TYisAKV (AVIS_TYPE (avis))) {
            new_avis = AvisCopyAsFunVardecs (avis, arg_info);
            INFO_FUN_AKV_IDS (arg_info)
                = TBmakeIds (new_avis, INFO_FUN_AKV_IDS (arg_info));
        } else {
            new_avis = AvisCopyAsFunArgs (avis, arg_info);
        }
    }
    
    DBUG_RETURN (arg_node);
}
node *
ATravIdsFun (node *arg_node, info *arg_info)
{
    node *avis, *new_avis;
    DBUG_ENTER ();
    
    avis = IDS_AVIS (arg_node);
    new_avis = LUTsearchInLutPp (INFO_LUT (arg_info), avis);
    // Not found in lut
    if (new_avis == avis) {
        new_avis = AvisCopyAsFunVardecs (avis, arg_info);
    }
    IDS_NEXT (arg_node) = TRAVopt (IDS_NEXT (arg_node), arg_info);

    DBUG_RETURN (arg_node);
}
/******************************************************************************
 *
 * function:
 *    node *LiftBodyToFoldopFun (char *fun_name, node *assigns, node *v1_avis, 
 *                               node *v2_avis, node *ret_avis, node *fundefs, 
 *                               info *arg_info)
 *
 * description:
 *    Lifts the given assigns into a separate function, using the ret_avis as
 *    the return value. Immediately calls dead code removal on the new function
 *    and removes unused arguments.
 *
 ******************************************************************************/
static node *
LiftBodyToFoldopFun (char *fun_name, node *assigns, node *v1_avis, node *v2_avis, node *ret_avis, node *fundefs, info *arg_info)
{
    node *new_v1_avis, *new_v2_avis;
    node *fundef, *fun_body;
    node *avis, *args;
    anontrav_t atrav_fun[3] = {{N_id, &ATravIdFun}, {N_ids, &ATravIdsFun}, {(nodetype)0, NULL}};
    
    DBUG_ENTER ();

    INFO_FUN_VARDECS (arg_info) = NULL;
    INFO_FUN_ARGS (arg_info) = NULL;
    
    assigns = TCappendAssign (
                    assigns, 
                    TBmakeAssign (
                        TBmakeReturn (TBmakeExprs (TBmakeId (ret_avis), NULL)), 
                        NULL));
    
    // V1 and V2 should be last args, ensure they are added first
    new_v2_avis = AvisCopyAsFunArgs (v2_avis, arg_info);
    new_v1_avis = AvisCopyAsFunArgs (v1_avis, arg_info);
    
    // Determine args and vardecs
    TRAVpushAnonymous (atrav_fun, &TRAVsons);
    assigns = TRAVopt (assigns, arg_info);
    TRAVpop ();
    
    // Copy body with LUT to avoid conflicts with avis and SSA
    fun_body = DUPdoDupTreeLut (assigns, INFO_LUT (arg_info));
    assigns = RemoveAssign (assigns, TCgetLastAssign (assigns));
    fun_body = RemoveAssign (fun_body, FindAssignLetPrf (fun_body, F_accu_dummy));
    
    // Prepend akv vardec initialisations
    while (INFO_FUN_AKV_IDS (arg_info) != NULL) {
        avis = IDS_AVIS (INFO_FUN_AKV_IDS (arg_info));
        fun_body = TBmakeAssign (
                       TBmakeLet (TBmakeIds (avis, NULL),
                                  COconstant2AST (TYgetValue (AVIS_TYPE (avis)))),
                       fun_body);
        AVIS_SSAASSIGN (avis) = fun_body;
        INFO_FUN_AKV_IDS (arg_info)
            = FREEdoFreeNode (INFO_FUN_AKV_IDS (arg_info));
    }
    
    // Create fundef
    fundef = TBmakeFundef (
        fun_name,
        NSdupNamespace (FUNDEF_NS (INFO_FUNDEF (arg_info))),
        TBmakeRet (TYcopyType (AVIS_TYPE (ret_avis)), NULL),
        INFO_FUN_ARGS (arg_info), 
        TBmakeBlock (fun_body, INFO_FUN_VARDECS (arg_info)), 
        fundefs);
    
    // Remove junk from function, break here
    fundef = DCRdoDeadCodeRemoval (fundef);
    
    // Remove unused arguments
    fundef = UAAdoUnusedArgumentAnnotate (fundef);
    
    ARG_ISUSEDINBODY (AVIS_DECL (new_v2_avis)) = TRUE;
    ARG_ISUSEDINBODY (AVIS_DECL (new_v1_avis)) = TRUE;
    
    // Can't use unused_function_argument_removal, as only for module, so do manually
    FUNDEF_ARGS (fundef) = RemoveUnusedArgs (FUNDEF_ARGS (fundef));
    
    // Set flags
    FUNDEF_ISGPUFUNCTION (fundef) = TRUE;
    FUNDEF_ISFOLDFUNCTION (fundef) = TRUE;
    RET_ISGPUMEM (FUNDEF_RETS (fundef)) = TRUE;
    args = FUNDEF_ARGS (fundef);
    while (args != NULL) {
        ARG_ISGPUMEM (args) = TRUE;
        args = ARG_NEXT (args);
    }
    
    DBUG_RETURN (fundef);
}
/******************************************************************************
 *
 * function:
 *    node *CalculateFoldopPartialArgs (node *fundef, info *arg_info)
 *
 * description:
 *    Calculates the arguments for the fundef in the with-loop bodies context. 
 *    The last 'to_exclude' number of variables are excluded.
 *
 ******************************************************************************/
static node *
CalculateFoldopPartialArgs (node *fundef, info *arg_info)
{
    node *fundef_ap_args, *fundef_ap_args_tmp;
    size_t partial_args_num;
    
    DBUG_ENTER ();
    
    partial_args_num = TCcountArgs (FUNDEF_ARGS (fundef));
    if (partial_args_num > 2) {
        // Get expression from args
        fundef_ap_args = TCcreateExprsFromArgs (FUNDEF_ARGS (fundef));
        fundef_ap_args_tmp = TCgetNthExprs (partial_args_num - 3, fundef_ap_args);
        EXPRS_NEXT (fundef_ap_args_tmp) = FREEoptFreeTree (EXPRS_NEXT (fundef_ap_args_tmp));
        // Use lut_inv to get variables from with loop context
        fundef_ap_args_tmp = fundef_ap_args;
        fundef_ap_args = DUPdoDupTreeLut (fundef_ap_args, INFO_LUT_INV (arg_info));
        fundef_ap_args_tmp = FREEdoFreeTree (fundef_ap_args_tmp);
    } else {
        fundef_ap_args = NULL;
    }
    
    DBUG_RETURN (fundef_ap_args);
}
/******************************************************************************
 *
 * function:
 *    node *CreateDeviceType (ntype *type)
 *
 * description:
 *    Creates a device version of the given type.
 *
 ******************************************************************************/
static ntype *
CreateDeviceType (ntype *type)
{
    ntype *device_type;
    simpletype fb_simpletype;
    DBUG_ENTER ();
    
    // This would be ideal, but we aren't allowed to do this :(
    //device_type = TYmakeAKD (TYmakeSimpleType (fb_simpletype), 1, SHcopyShape (TYgetShape (fb_val_type)));
    
    //fb_simpletype = CUh2dSimpleTypeConversion (TYgetSimpleType (TYgetScalar (AVIS_TYPE (neutr_avis))));
    //device_type = TYeliminateAKV (AVIS_TYPE (neutr_avis));
    //device_type = TYsetScalar (device_type, TYmakeSimpleType (fb_simpletype));
    fb_simpletype = CUh2dSimpleTypeConversion (TYgetSimpleType (TYgetScalar (type)));
    device_type = TYsetScalar (TYeliminateAKV (type), TYmakeSimpleType (fb_simpletype)); 
    
    DBUG_RETURN (device_type);
}
/******************************************************************************
 *
 * function:
 *    node *CreateFoldbufferType (ntype *device_type, int size)
 *
 * description:
 *    Creates a foldbuffer for the given device type, does not consume the
 *    device type.
 *
 ******************************************************************************/
static ntype *
CreateFoldbufferType (ntype *device_type, int size)
{
    shape *shp;
    DBUG_ENTER ();
    // device_type is always AKS
    if (size > 0) {
        shp = SHcreateShape (1, size);
        device_type = TYmakeAKS (TYmakeSimpleType (TYgetSimpleType (TYgetScalar (device_type))), 
                                 SHappendShapes (shp, TYgetShape (device_type)));
        shp = SHfreeShape (shp);
    } else {
        device_type = TYmakeAKD (TYcopyType (TYgetScalar (device_type)), 
                             (size_t)(1 + TYgetDim (device_type)), 
                             SHmakeShape (0));
    }
    DBUG_RETURN (device_type);
}
/******************************************************************************
 *
 * function:
 *    node *CreateFoldbufferAssign (node *val_avis, node* size_expr, node *let_lhs_avis, info *arg_info)
 *
 * description:
 *    Creates a foldbuffer avis and assign node, returns the avis. Consumes
 *    the size_expr.
 *
 ******************************************************************************/
static node *
CreateFoldbufferAssign (node *val_avis, node* size_expr, node *let_lhs_avis, info *arg_info)
{
    node *fb_avis, *fb_expr;
    ntype *device_type;
    DBUG_ENTER ();
    
    /* Use TBmakeStr to ensure CUFD_noops from the same with-loop are not
     * alliased. They need to be seperate memory.
     */
    device_type = CreateDeviceType (AVIS_TYPE (val_avis));
    fb_expr = TCmakePrf3 (F_CUFD_noop, // CUFD_noop (size, type, lhs name)
                          size_expr,
                          TBmakeType (device_type),
                          TBmakeStr (STRcpy (AVIS_NAME (let_lhs_avis))));
    
    device_type = CreateFoldbufferType (device_type, 0);
    INFO_PREWITH_ASSINGS (arg_info)
        = AppendLetAssign ("foldbuf", 
                           device_type,
                           fb_expr,
                           INFO_PREWITH_ASSINGS (arg_info), 
                           arg_info);
    fb_avis = GetLastIdsAvis (INFO_PREWITH_ASSINGS (arg_info));
    
    DBUG_RETURN (fb_avis);
}
/******************************************************************************
 *
 * function:
 *    node *CreateZeroAssign (info *arg_info)
 *
 * description:
 *    Creates an avis with the value 0.
 *
 ******************************************************************************/
static node *
CreateZeroAssign (info *arg_info)
{
    node *zero_avis;
    DBUG_ENTER ();
    
    INFO_PREWITH_ASSINGS (arg_info)
        = AppendLetAssign ("const", 
                           CreateScalarType (T_int), 
                           TBmakeNum (0), 
                           INFO_PREWITH_ASSINGS (arg_info), 
                           arg_info);
    zero_avis = GetLastIdsAvis (INFO_PREWITH_ASSINGS (arg_info));
    
    DBUG_RETURN (zero_avis);
}
/******************************************************************************
 *
 * function:
 *    node *CreateSpaceAssign (info *arg_info) 
 *
 * description:
 *    Creates a space avis and assign node, returns the avis. The space
 *    represents the index space of the given partition.
 *    The RHS is given the following form:
 *        space = ((ub - lb) / s) * w;
 *
 ******************************************************************************/
static node *
CreateSpaceAssign (node *wl_part, info *arg_info) 
{
    node *b1_expr, *b2_expr;
    node *b1_avis, *b2_avis;
    node *step_avis, *step_expr;
    node *step_minus_avis, *step_minus_expr;
    node *width_avis, *width_expr;
    node *space_avis, *space_expr;
    ntype *generic_type;
    
    DBUG_ENTER ();
    // int[*]
    generic_type = TYmakeAUD (TYmakeSimpleType (T_int));
    
    // int[.] b1 = ...;
    if (NODE_TYPE (PART_BOUND1 (wl_part)) == N_id) {
        b1_expr = DUPdoDupTree (PART_BOUND1 (wl_part));
    } else {
        INFO_PREWITH_ASSINGS (arg_info)
            = AppendLetAssign ("bound1", 
                               NTCnewTypeCheck_Expr (PART_BOUND1 (wl_part)), 
                               DUPdoDupTree (PART_BOUND1 (wl_part)), 
                               INFO_PREWITH_ASSINGS (arg_info), 
                               arg_info);
        b1_avis = GetLastIdsAvis (INFO_PREWITH_ASSINGS (arg_info));
        b1_expr = TBmakeId (b1_avis);
    }
    // int[.] b2 = ...;
    if (NODE_TYPE (PART_BOUND2 (wl_part)) == N_id) {
        b2_expr = DUPdoDupTree (PART_BOUND2 (wl_part));
    } else {
        INFO_PREWITH_ASSINGS (arg_info)
            = AppendLetAssign ("bound2", 
                               NTCnewTypeCheck_Expr (PART_BOUND2 (wl_part)), 
                               DUPdoDupTree (PART_BOUND2 (wl_part)), 
                               INFO_PREWITH_ASSINGS (arg_info), 
                               arg_info);
        b2_avis = GetLastIdsAvis (INFO_PREWITH_ASSINGS (arg_info));
        b2_expr = TBmakeId (b2_avis);
    }
    
    // int[.] space = b2 - b1;
    space_expr = TCmakePrf2 (F_sub_VxV, b2_expr, b1_expr);
    INFO_PREWITH_ASSINGS (arg_info)
        = AppendLetAssign ("space", 
                           generic_type, 
                           space_expr, 
                           INFO_PREWITH_ASSINGS (arg_info), 
                           arg_info);
    space_avis = GetLastIdsAvis (INFO_PREWITH_ASSINGS (arg_info));
    
    if (PART_STEP (wl_part) != NULL) {
        // int[.] space'' = (space + step - 1) / step;
        if (NODE_TYPE (PART_STEP (wl_part)) == N_id) {
            step_avis = ID_AVIS (PART_STEP (wl_part));
            step_expr = DUPdoDupTree (PART_STEP (wl_part));
        } else {
            INFO_PREWITH_ASSINGS (arg_info)
                = AppendLetAssign ("step", 
                                   NTCnewTypeCheck_Expr (PART_STEP (wl_part)), 
                                   DUPdoDupTree (PART_STEP (wl_part)), 
                                   INFO_PREWITH_ASSINGS (arg_info), 
                                   arg_info);
            step_avis = GetLastIdsAvis (INFO_PREWITH_ASSINGS (arg_info));
            step_expr = TBmakeId (step_avis);
        }
        
        // int[.] step_minus = step - 1;
        step_minus_expr = TCmakePrf2 (F_sub_VxS, step_expr, TBmakeNum (1));
        INFO_PREWITH_ASSINGS (arg_info)
            = AppendLetAssign ("step_minus", 
                               TYeliminateAKV (AVIS_TYPE (step_avis)),
                               step_minus_expr, 
                               INFO_PREWITH_ASSINGS (arg_info), 
                               arg_info);
        step_minus_avis = GetLastIdsAvis (INFO_PREWITH_ASSINGS (arg_info));
        
        // int[.] space' = space + step_minus;
        space_expr = TCmakePrf2 (F_add_VxV, TBmakeId (space_avis), TBmakeId (step_minus_avis));
        INFO_PREWITH_ASSINGS (arg_info)
            = AppendLetAssign ("space", 
                               TYcopyType (generic_type),
                               space_expr, 
                               INFO_PREWITH_ASSINGS (arg_info), 
                               arg_info);
        space_avis = GetLastIdsAvis (INFO_PREWITH_ASSINGS (arg_info));
        
        // int[.] space'' = space' / step;
        space_expr = TCmakePrf2 (F_div_VxV, TBmakeId (space_avis), DUPdoDupTree (step_expr));
        INFO_PREWITH_ASSINGS (arg_info)
            = AppendLetAssign ("space", 
                               TYcopyType (generic_type),
                               space_expr, 
                               INFO_PREWITH_ASSINGS (arg_info), 
                               arg_info);
        space_avis = GetLastIdsAvis (INFO_PREWITH_ASSINGS (arg_info));
        if (PART_WIDTH (wl_part) != NULL) {
            if (NODE_TYPE (PART_WIDTH (wl_part)) == N_id) {
                width_expr = DUPdoDupTree (PART_WIDTH (wl_part));
            } else {
                INFO_PREWITH_ASSINGS (arg_info)
                    = AppendLetAssign ("width", 
                                       NTCnewTypeCheck_Expr (PART_WIDTH (wl_part)), 
                                       DUPdoDupTree (PART_WIDTH (wl_part)), 
                                       INFO_PREWITH_ASSINGS (arg_info), 
                                       arg_info);
                width_avis = GetLastIdsAvis (INFO_PREWITH_ASSINGS (arg_info));
                width_expr = TBmakeId (width_avis);
            }
            // int[.] space''' = space'' * width;
            space_expr = TCmakePrf2 (F_mul_VxV, TBmakeId (space_avis), width_expr);
            INFO_PREWITH_ASSINGS (arg_info)
                = AppendLetAssign ("space", 
                                   TYcopyType (generic_type),
                                   space_expr, 
                                   INFO_PREWITH_ASSINGS (arg_info), 
                                   arg_info);
            space_avis = GetLastIdsAvis (INFO_PREWITH_ASSINGS (arg_info));
        }
    }
    DBUG_RETURN (space_avis);
}
/******************************************************************************
 *
 * function:
 *    node *CreateElemsAssign (node *space_avis, info *arg_info)
 *
 * description:
 *    Creates a elems avis and assign node, returns the avis. The elems
 *    represents the total amount of indices called.
 *    The RHS is given the following form:
 *        elems = prod (space);
 *
 ******************************************************************************/
static node *
CreateElemsAssign (node *space_avis, info *arg_info)
{
    node *elems_avis, *elems_expr;
    DBUG_ENTER ();
    
    // Uses "int prod (int[*])"
    // int elems = prod (space);
    elems_expr = TBmakeId (space_avis);
    elems_expr = DSdispatchFunCall (NSgetNamespace (global.preludename), "prod",
                                    TBmakeExprs (elems_expr, NULL));
    DBUG_ASSERT (elems_expr != NULL, "%s::prod not found", global.preludename);
    
    INFO_PREWITH_ASSINGS (arg_info)
        = AppendLetAssign ("elems", 
                           CreateScalarType (T_int),
                           elems_expr, 
                           INFO_PREWITH_ASSINGS (arg_info), 
                           arg_info);
    elems_avis = GetLastIdsAvis (INFO_PREWITH_ASSINGS (arg_info));
    
    DBUG_RETURN (elems_avis);
}
/******************************************************************************
 *
 * function:
 *    node *CreateGridAssign (node *elems_avis, info *arg_info)
 *
 * description:
 *    Creates a grid avis and assign node, returns the avis. The grid
 *    represents the size of the eventual grid for these elements.
 *    The RHS is given the following form:
 *        grid = (elems + 1023) / 1024;
 *
 ******************************************************************************/
static node *
CreateGridAssign (node *elems_avis, info *arg_info)
{
    node *grid_avis, *grid_expr;
    DBUG_ENTER ();
    
    // int gird = elems + 1023
    grid_expr = TCmakePrf2 (F_add_SxS, TBmakeId (elems_avis), TBmakeNum (1023));
    INFO_PREWITH_ASSINGS (arg_info)
        = AppendLetAssign ("grid", 
                           CreateScalarType (T_int),
                           grid_expr, 
                           INFO_PREWITH_ASSINGS (arg_info), 
                           arg_info);
    grid_avis = GetLastIdsAvis (INFO_PREWITH_ASSINGS (arg_info));
    
    // int gird' = grid / 1024
    grid_expr = TCmakePrf2 (F_div_SxS, TBmakeId (grid_avis), TBmakeNum (1024));
    INFO_PREWITH_ASSINGS (arg_info)
        = AppendLetAssign ("grid", 
                           CreateScalarType (T_int),
                           grid_expr, 
                           INFO_PREWITH_ASSINGS (arg_info), 
                           arg_info);
    grid_avis = GetLastIdsAvis (INFO_PREWITH_ASSINGS (arg_info));
    
    DBUG_RETURN (grid_avis);
}
/******************************************************************************
 *
 * function:
 *    node *CreateFinalUbAssign (node *grid_avis, info *arg_info)
 *
 * description:
 *    Creates a final_ub avis and assign node, returns the avis. The final_ub
 *    will be the upper bound used in the final fold.
 *    The RHS is given the following form:
 *        final_ub = min (grid, 1024);
 *
 ******************************************************************************/
static node *
CreateFinalUbAssign (node *grid_avis, info *arg_info)
{
    node *final_ub_avis, *final_ub_expr;
    DBUG_ENTER ();
    
    // int final_ub = min (grid, 1024);
    final_ub_expr = TCmakePrf2 (F_min_SxS, 
                                TBmakeId (grid_avis), 
                                TBmakeNum (1024));
    INFO_PREWITH_ASSINGS (arg_info)
        = AppendLetAssign ("final_ub", 
                           CreateScalarType (T_int),
                           final_ub_expr, 
                           INFO_PREWITH_ASSINGS (arg_info), 
                           arg_info);
    final_ub_avis = GetLastIdsAvis (INFO_PREWITH_ASSINGS (arg_info));
    
    DBUG_RETURN (final_ub_avis);
}
/******************************************************************************
 *
 * function:
 *    node *CreateTXAssign (node *wl_body, info *arg_info)
 *
 * description:
 *    Creates and avis and assign for retrieving the cuda thread X index. The
 *    avis is stored in INFO_TX_AVIS. The assignment is added to the provided 
 *    wl_body, which is returned afterwards. 
 *
 ******************************************************************************/
static node *
CreateTXAssign (node *wl_body, info *arg_info)
{
    DBUG_ENTER ();
    
    wl_body = PreppendLetAssign ("tx", 
                                 CreateScalarType (T_int),
                                 TBmakePrf (F_cuda_threadIdx_x, NULL),
                                 wl_body, 
                                 arg_info);
    INFO_TX_AVIS (arg_info) = GetFirstIdsAvis (wl_body);
    
    DBUG_RETURN (wl_body);
}
/******************************************************************************
 *
 * function:
 *    node *CreateOffsetAssign (node *wl_body, node *index_avis, int val_size, info *arg_info)
 *
 * description:
 *    Creates and avis and assign to calculate the tx offset. This is done when
 *    the values of the fold are non-scalars. It simply adds the size of the
 *    value shape to the tx. 
 *
 ******************************************************************************/
static node *
CreateTXOffsetAssign (node *wl_body, node *tx_avis, int val_size, info *arg_info)
{
    DBUG_ENTER ();
    
    wl_body 
        = PreppendLetAssign ("tx_offset", 
                             CreateScalarType (T_int),
                             TCmakePrf2 (F_mul_SxS,
                                         TBmakeId (tx_avis),
                                         TBmakeNum (val_size)),
                             wl_body, 
                             arg_info);
    
    DBUG_RETURN (wl_body);
}
/******************************************************************************
 *
 * function:
 *    node *CreateBXAssign (node *wl_body, info *arg_info)
 *
 * description:
 *    Creates and avis and assign for retrieving the cuda thread-block X index.
 *    The assignment is added to the provided wl_body.
 * description:
 *    Creates and avis and assign for retrieving the cuda thread-block X index. 
 *    The avis is stored in INFO_BX_AVIS. The assignment is added to the 
 *    provided wl_body, which is returned afterwards. 
 *
 ******************************************************************************/
static node *
CreateBXAssign (node *wl_body, info *arg_info)
{
    DBUG_ENTER ();
    
    wl_body = PreppendLetAssign ("bx", 
                                 CreateScalarType (T_int),
                                 TBmakePrf (F_cuda_blockIdx_x, NULL),
                                 wl_body, 
                                 arg_info);
    INFO_BX_AVIS (arg_info) = GetFirstIdsAvis (wl_body);
    
    DBUG_RETURN (wl_body);
}
/******************************************************************************
 *
 * function:
 *    node *CreateCondAssign (node *wl_body, node *tx_avis, info *arg_info)
 *
 * description:
 *    Creates and avis and assign to check if tx == 0. The assignment is added 
 *    to the provided wl_body, which is returned afterwards. 
 *
 ******************************************************************************/
static node *
CreateCondAssign (node *wl_body, node *tx_avis, info *arg_info)
{
    DBUG_ENTER ();
    
    wl_body = AppendLetAssign ("cond", 
                               CreateScalarType (T_bool),
                               TCmakePrf2 (F_eq_SxS, TBmakeId (tx_avis), TBmakeNum (0)),
                               wl_body, 
                               arg_info);
    
    DBUG_RETURN (wl_body);
}
/******************************************************************************
 *
 * function:
 *    node *CreatePrefoldStartAssign (node *wl_body, node *tx_avis, info *arg_info)
 *
 * description:
 *    Creates an offset avis equal to tx_avis + 1024. 
 *
 ******************************************************************************/
static node *
CreatePrefoldStartAssign (node *wl_body, node *tx_avis, info *arg_info)
{
    DBUG_ENTER ();
    
    wl_body = PreppendLetAssign ("prefold_start", 
                                 CreateScalarType (T_int),
                                 TCmakePrf2 (F_add_SxS,
                                             TBmakeId (tx_avis), 
                                             TBmakeNum (1024)),
                                 wl_body, 
                                 arg_info);
    INFO_FINAL_OFFSET_START_AVIS (arg_info) = GetFirstIdsAvis (wl_body);
    
    DBUG_RETURN (wl_body);
}
/******************************************************************************
 *
 * function:
 *    node *GetOrCreateNeutrAvis (node *fold_node, info *arg_info)
 *
 * description:
 *    Checks if the neutral element of the provided fold is an id. If it is, it
 *    will retrieve the associated avis. If not, it lifts it the expression 
 *    into a separate vardec and assign and returns the new avis.
 *
 ******************************************************************************/
static node *
GetOrCreateNeutrAvis (node *fold_node, info *arg_info)
{
    node *neutr_avis;
    DBUG_ENTER ();
    
    // Lift neutral value to id if it isn't already
    if (NODE_TYPE (FOLD_NEUTRAL (fold_node)) == N_id) {
        neutr_avis = ID_AVIS (FOLD_NEUTRAL (fold_node));
    } else {
        INFO_PREWITH_ASSINGS (arg_info) 
            = PreppendLetAssign ("neutr", 
                                 NTCnewTypeCheck_Expr (FOLD_NEUTRAL (fold_node)),
                                 FOLD_NEUTRAL (fold_node),
                                 INFO_PREWITH_ASSINGS (arg_info), 
                                 arg_info);
        neutr_avis = GetFirstIdsAvis (INFO_PREWITH_ASSINGS (arg_info));
        FOLD_NEUTRAL (fold_node) = TBmakeId (neutr_avis);
    }

    DBUG_RETURN (neutr_avis);
}
/******************************************************************************
 *
 * function:
 *    node *CreateSdataAssign (node *neutr_avis, node *fold, node* assign_chain, int val_size, node *let_lhs_avis, info *arg_info)
 *
 * description:
 *    Creates and avis and assign to allocate the folding shared memory.
 *    Returns the created assign.
 *
 ******************************************************************************/
static node *
CreateSdataAssign (node *neutr_avis, node *fold, node* assign_chain, int val_size, node *let_lhs_avis, info *arg_info)
{
    simpletype sdata_simpletype;
    shape *sdata_inner_shape;
    ntype *sdata_type;
    node *sdata_expr, *sdata_assign;
    DBUG_ENTER ();
    
    // Calculate type
    sdata_simpletype = CUh2shSimpleTypeConversion (
                               TYgetSimpleType (TYgetScalar (AVIS_TYPE (neutr_avis))));
    sdata_inner_shape = SHcreateShape (1, 1024);
    sdata_type = TYmakeAKS (TYmakeSimpleType (sdata_simpletype), 
                            SHappendShapes (sdata_inner_shape, 
                                            TYgetShape (AVIS_TYPE (neutr_avis))));
    SHfreeShape (sdata_inner_shape);
    
    // Create assign
    sdata_expr = TCmakePrf4 (F_init_shared,
                             TBmakeType (sdata_type),
                             TBmakeId (neutr_avis),
                             TBmakeNum (val_size),
                             TBmakeStr (STRcpy (AVIS_NAME (let_lhs_avis))));
    sdata_assign = PreppendLetAssign ("sdata", 
                                      TYcopyType (sdata_type), 
                                      sdata_expr, 
                                      assign_chain, 
                                      arg_info);
    AVIS_SHAREDINITVALUE (GetFirstIdsAvis (sdata_assign)) = sdata_expr;
    
    DBUG_RETURN (sdata_assign);
}
/******************************************************************************
 *
 * function:
 *    node *CreateSdataModarrayAssign (node *sdata_avis, node *index_avis, node *val_avis, node *wl_body, info *arg_info) 
 *
 * description:
 *    Creates and avis and assign to assign the val_avis to the sdata_avis at
 *    the index_avis. Returns wl_body with the assign appended.
 *
 ******************************************************************************/
static node *
CreateSdataModarrayAssign (node *sdata_avis, node *index_avis, node *val_avis, node *wl_body, info *arg_info) 
{
    node *new_sdata_avis, *sdata_assign;
    prf prf;
    
    DBUG_ENTER ();
    new_sdata_avis = CopyAvisAndVardec (sdata_avis, arg_info);
    prf = TUisScalar (AVIS_TYPE (val_avis)) ? F_idx_modarray_AxSxS : F_idx_modarray_AxSxA;
    sdata_assign = MakeAssignLet (
                       TBmakeIds (new_sdata_avis, NULL),
                       TCmakePrf3 (prf,
                                   TBmakeId (sdata_avis),
                                   TBmakeId (index_avis),
                                   TBmakeId (val_avis)),
                       NULL);
    wl_body = TCappendAssign (wl_body, sdata_assign);
    
    DBUG_RETURN (wl_body);
}
/******************************************************************************
 *
 * function:
 *    node *CreateFoldPrfAssign (prf prf, node *prf_exprs, node *sdata_avis, node *foldop, node *next_assign info *arg_info)
 *
 * description:
 *    Creates and avis and assign to for the given prf, sets PRF_FOLDOP.
 *    Returns the created assign.
 *
 ******************************************************************************/
static node *
CreateFoldPrfAssign (prf prf, node *prf_exprs, node *sdata_avis, node *foldop, node *next_assign, info *arg_info)
{
    // tx, from, to, fb, shmem)
    node *prf_expr, *lhs_avis;
    
    DBUG_ENTER ();
    
    prf_expr = TBmakePrf (prf, prf_exprs);
    PRF_FOLDOP (prf_expr) = foldop;
    next_assign = MakeAssignLet (NULL, prf_expr, next_assign);
    lhs_avis = CopyAvisAndVardec (sdata_avis, arg_info);
    ASSIGN_LHS (next_assign)= TBmakeIds (lhs_avis, NULL);
    
    DBUG_RETURN (next_assign);
}
/******************************************************************************
 *
 * function:
 *    node *CreateFinalFoldEmptyWithAssign (node *ub_avis, info *arg_info)
 *
 * description:
 *    Creates a with-loop with a single code block and partition ranging from 0
 *    to ub_avis.
 *
 ******************************************************************************/
static node *
CreateFinalFoldEmptyWithAssign (node *ub_avis, info *arg_info)
{
    node *with, *code, *part, *withid, *generator;
    node *lb_array, *ub_array;
    node *vec_avis, *ids_avis;
    node *assign;
    DBUG_ENTER ();
    
    lb_array = TCmakeIntVector (TCmakeExprsNum (0));
    ub_array = TCmakeIntVector (TBmakeExprs (TBmakeId (ub_avis), NULL));
    generator = TBmakeGenerator (F_wl_le, F_wl_lt, lb_array, ub_array, NULL, NULL);
    ids_avis = MakeAvisAndVardec (TRAVtmpVarName ("i"), 
                                TYmakeAKS (TYmakeSimpleType (T_int), SHcreateShape (0)), 
                                arg_info);
    vec_avis = MakeAvisAndVardec (TRAVtmpVarName ("vec"), 
                                 TYmakeAKS (TYmakeSimpleType (T_int), SHcreateShape (1, 1)), 
                                 arg_info);
    withid = TBmakeWithid (TBmakeIds (vec_avis, NULL), TBmakeIds (ids_avis, NULL));
    code = TBmakeCode (MAKE_EMPTY_BLOCK(), NULL);
    CODE_USED (code) = 1;
    part = TBmakePart (code, withid, generator);
    with = TBmakeWith (part, code, NULL);
    WITH_CUDARIZABLE (with) = TRUE;
    assign = MakeAssignLet (NULL, with, NULL);
    INFO_FINAL_INDEX_AVIS (arg_info) = ids_avis;
    
    DBUG_RETURN (assign);
}

/** <!--********************************************************************-->
 * @}  <!-- Static helper functions -->
 *****************************************************************************/
 
/** <!--********************************************************************-->
 *
 * @name Traversal functions
 * @{
 *
 *****************************************************************************/
node *
CUFDmodule (node *arg_node, info *arg_info)
{
    DBUG_ENTER ();

    DSinitDeserialize (arg_node);
    arg_node = TRAVsons (arg_node, arg_info);
    DSfinishDeserialize (arg_node);

    DBUG_RETURN (arg_node);
}

node *
CUFDfundef (node *arg_node, info *arg_info)
{
    DBUG_ENTER ();
    
    // Generate GPU fun and add fold prf
    INFO_FUNDEF (arg_info) = arg_node;
    FUNDEF_BODY (arg_node) = TRAVopt (FUNDEF_BODY (arg_node), arg_info);
    INFO_FUNDEF (arg_info) = NULL;
    
    if (FUNDEF_NEXT (arg_node) == NULL) {
        // Append gpu functions
        FUNDEF_NEXT (arg_node) = INFO_GPU_FUNS (arg_info);
        INFO_GPU_FUNS (arg_info) = NULL;
    } else {
        FUNDEF_NEXT (arg_node) = TRAVdo (FUNDEF_NEXT (arg_node), arg_info);
    }

    DBUG_RETURN (arg_node);
}

node *
CUFDassign (node *arg_node, info *arg_info)
{
    node *prewith_assigns, *postwith_assigns;
    DBUG_ENTER ();
    
    INFO_POSTWITH_ASSINGS (arg_info) = NULL;
    INFO_PREWITH_ASSINGS (arg_info) = NULL;
    INFO_ASSIGN (arg_info) = arg_node;
    ASSIGN_STMT (arg_node) = TRAVopt (ASSIGN_STMT (arg_node), arg_info);
    postwith_assigns = INFO_POSTWITH_ASSINGS (arg_info);
    prewith_assigns = INFO_PREWITH_ASSINGS (arg_info);
    
    ASSIGN_NEXT (arg_node) = TRAVopt (ASSIGN_NEXT (arg_node), arg_info);
    
    postwith_assigns = TCappendAssign (postwith_assigns, ASSIGN_NEXT (arg_node));
    ASSIGN_NEXT (arg_node) = postwith_assigns;
    arg_node = TCappendAssign (prewith_assigns, arg_node);  
    
    DBUG_RETURN (arg_node);
}

node *
CUFDlet (node *arg_node, info* arg_info)
{
    DBUG_ENTER ();
    
    INFO_LET_LHS (arg_info) = LET_IDS (arg_node);
    LET_EXPR (arg_node) = TRAVopt (LET_EXPR (arg_node), arg_info);
    INFO_LET_LHS (arg_info) = NULL;
    
    DBUG_RETURN (arg_node);
}

node *
CUFDwith (node *arg_node, info *arg_info)
{
    node *space_avis, *elems_avis, *final_ub_avis;
    DBUG_ENTER ();
    
    if (WITH_TYPE (arg_node) == N_fold)
    {
        INFO_WITH (arg_info) = arg_node;
        if(WITH_CUDARIZABLE (arg_node))
        {
            DBUG_PRINT ("Found cudarizable fold in \"%s\"", 
                        FUNDEF_NAME (INFO_FUNDEF (arg_info)));
            
            // Create prewith assignments
            DBUG_PRINT ("Creating initial pre-with assignments");
            space_avis = CreateSpaceAssign (WITH_PART (INFO_WITH (arg_info)), arg_info);
            elems_avis = CreateElemsAssign (space_avis, arg_info);
            INFO_GRID_AVIS (arg_info) = CreateGridAssign (elems_avis, arg_info);
            final_ub_avis = CreateFinalUbAssign (INFO_GRID_AVIS (arg_info), arg_info);
            // Create post with assignments
            DBUG_PRINT ("Creating initial post-with assignments");
            INFO_POSTWITH_ASSINGS (arg_info) = CreateFinalFoldEmptyWithAssign (final_ub_avis, arg_info);
            INFO_FINAL_WITH_ASSIGN (arg_info) = INFO_POSTWITH_ASSINGS (arg_info);
            INFO_FINAL_WITH (arg_info) = ASSIGN_RHS (INFO_FINAL_WITH_ASSIGN (arg_info));
            
            // Create GPU function and update with loop body
            INFO_FIRST_FOLD (arg_info) = TRUE;
            WITH_WITHOP (arg_node) = TRAVdo (WITH_WITHOP (arg_node), arg_info);
            
            // Set pragma to NULL
            if (PART_PRAGMA (WITH_PART (arg_node)) != NULL) {
                CTIwarn (LINE_TO_LOC (global.linenum),
                         "Cudarizable fold with-loops do not support user defined pragmas!");
                PART_PRAGMA (WITH_PART (arg_node))
                    = FREEdoFreeTree (PART_PRAGMA (WITH_PART (arg_node)));
            }
            
        } else { 
            DBUG_PRINT ("Found non-cudarizable fold in \"%s\"", 
                        FUNDEF_NAME (INFO_FUNDEF (arg_info)));
            
            WITH_CODE (arg_node) = TRAVopt (WITH_CODE (arg_node), arg_info);
        }
        INFO_WITH (arg_info) = NULL;
    } else {
        // Have to traverse code bodies to find any secret nested with-loops
        WITH_CODE (arg_node) = TRAVopt (WITH_CODE (arg_node), arg_info);
    }
    
    DBUG_RETURN (arg_node);
}

node *
CUFDfold (node *arg_node, info *arg_info)
{
    node *wl_body, *wl2_body;
    node *cexpr_avis, *gpu_res_avis, *neutr_avis;
    node *sdata_avis, *fb_avis;
    node *sdata2_avis, *fb2_avis;
    node *val_avis;
    node *foldop_fundef, *foldop_args;
    node *v1_avis, *v2_avis;
    node *val2_avis, *cexpr2_avis;
    node *neutr_arr_expr;
    node *new_fold;
    node *let_lhs_avis;
    node *tx_offset_avis, *i_offset_avis;
    bool is_scl;
    ntype *cexpr_type, *device_type;
    int val_size;
    
    DBUG_ENTER ();
    
    // Global data
    wl_body = CODE_CBLOCK_ASSIGNS (WITH_CODE (INFO_WITH (arg_info)));
    wl2_body = CODE_CBLOCK_ASSIGNS (WITH_CODE (INFO_FINAL_WITH (arg_info)));
    
    // First fold traversal, set arg_info
    if (INFO_FIRST_FOLD (arg_info)) {
        INFO_FIRST_FOLD (arg_info) = FALSE;
        INFO_SDATA_ASSIGNS (arg_info) = NULL;
        INFO_CUDA_FOLD_ASSIGNS (arg_info) = NULL;
        INFO_RES_ASSIGNS (arg_info) = NULL;
        INFO_FINAL_SDATA_ASSIGNS (arg_info) = NULL;
        INFO_FINAL_PREFOLD_ASSIGNS (arg_info) = NULL;
        INFO_FINAL_CUDA_FOLD_ASSIGNS (arg_info) = NULL;
        INFO_FINAL_RES_ASSIGNS (arg_info) = NULL;
        
        // Calculate val_size
        
        // Create LUTs
        INFO_LUT (arg_info) = LUTgenerateLut ();
        INFO_LUT_INV (arg_info) = LUTgenerateLut ();
        
        DBUG_PRINT ("Traversing first fold for with-loop");
        /// First with
        DBUG_PRINT ("  Removing accu");
        wl_body = RemoveAssign (wl_body, FindAssignLetPrf (wl_body, F_accu));
        DBUG_PRINT ("  Retrieving accu_dummy ids");
        INFO_ACCU_DUMMY_IDS (arg_info) 
            = ASSIGN_LHS (FindAssignLetPrf (wl_body, F_accu_dummy));
        DBUG_PRINT ("  Retrieving cexpr ids");
        INFO_CEXPRS (arg_info) = WITH_CEXPRS (INFO_WITH (arg_info));
        DBUG_PRINT ("  Creating tx allocation");
        wl_body = CreateTXAssign (wl_body, arg_info);
        DBUG_PRINT ("  Creating bx allocation");
        wl_body = CreateBXAssign (wl_body, arg_info);
        DBUG_PRINT ("  Creating cond allocation");
        wl_body = CreateCondAssign (wl_body, INFO_TX_AVIS (arg_info), arg_info);
        INFO_COND_AVIS (arg_info) = GetLastIdsAvis (wl_body);
        
        /// Second with
        DBUG_PRINT ("  Creating final cond allocation");
        wl2_body = CreateCondAssign (wl2_body, INFO_FINAL_INDEX_AVIS (arg_info), arg_info);
        INFO_FINAL_COND_AVIS (arg_info) = GetLastIdsAvis (wl2_body);
        
        DBUG_PRINT ("  Creating prefold start allocation");
        wl2_body = CreatePrefoldStartAssign (wl2_body, INFO_FINAL_INDEX_AVIS (arg_info), arg_info);
    
        // Add prep for second wl
        DBUG_PRINT ("Finished first fold for with-loop");
        CODE_CBLOCK_ASSIGNS (WITH_CODE (INFO_WITH (arg_info))) = wl_body;
        CODE_CBLOCK_ASSIGNS (WITH_CODE (INFO_FINAL_WITH (arg_info))) = wl2_body;
    }
    
    // Check if cexpr and dummy ids match amount of folds
    DBUG_ASSERT (INFO_CEXPRS (arg_info) != NULL,
                 "  Amount of cexprs does not match amount of folds");
    DBUG_ASSERT (INFO_ACCU_DUMMY_IDS (arg_info) != NULL,
                 "  Amount of dummy ids does not match amount of folds");
    
    DBUG_PRINT ("Starting standard fold traversal");
    // Retrieve accu_sel variables
    cexpr_avis = ID_AVIS (EXPRS_EXPR1 (INFO_CEXPRS (arg_info)));
    gpu_res_avis = ID_AVIS (PRF_ARG2 (ASSIGN_RHS (AVIS_SSAASSIGN (cexpr_avis))));
    val_avis = ID_AVIS (PRF_ARG3 (ASSIGN_RHS (AVIS_SSAASSIGN (cexpr_avis))));
    neutr_avis = GetOrCreateNeutrAvis (arg_node, arg_info);
    let_lhs_avis = IDS_AVIS (INFO_LET_LHS (arg_info));
    
    DBUG_PRINT ("  Generating pre-with foldbuf allocation");
    // Create allocated, but uninitialised buffer for partial results
    fb_avis = CreateFoldbufferAssign (val_avis, 
                                      TBmakeId (INFO_GRID_AVIS (arg_info)), 
                                      let_lhs_avis,
                                      arg_info);
    FOLD_PARTIALMEM (arg_node) = TBmakeId (fb_avis);
    
    DBUG_PRINT ("  Generating non-scalar allocations");
    // Handle non-scalar stuff
    cexpr_type = AVIS_TYPE (cexpr_avis);
    is_scl = TUisScalar (cexpr_type);
    if (is_scl) {
        val_size = 1;
        tx_offset_avis = INFO_TX_AVIS (arg_info);
        i_offset_avis = INFO_FINAL_INDEX_AVIS (arg_info);
    } else {
        val_size = TUgetLengthEncoding (cexpr_type);
        // Append to tx assign
        ASSIGN_NEXT (AVIS_SSAASSIGN (INFO_TX_AVIS (arg_info)))
            = CreateTXOffsetAssign (ASSIGN_NEXT (AVIS_SSAASSIGN (INFO_TX_AVIS (arg_info))), 
                                    INFO_TX_AVIS (arg_info), 
                                    val_size, 
                                    arg_info);
        tx_offset_avis = GetFirstIdsAvis (ASSIGN_NEXT (AVIS_SSAASSIGN (INFO_TX_AVIS (arg_info))));
        // Prepend to wl body
        wl2_body = CreateTXOffsetAssign (wl2_body, 
                                         INFO_FINAL_INDEX_AVIS (arg_info), 
                                         val_size,
                                         arg_info);
        i_offset_avis = GetFirstIdsAvis (wl2_body);
    }
    
    DBUG_PRINT ("  Generating GPU foldop function");
    // Retrieve v1 and v2 dummy values
    v1_avis = IDS_AVIS (INFO_ACCU_DUMMY_IDS (arg_info));
    INFO_ACCU_DUMMY_IDS (arg_info) = IDS_NEXT (INFO_ACCU_DUMMY_IDS (arg_info));
    v2_avis = IDS_AVIS (INFO_ACCU_DUMMY_IDS (arg_info));
    INFO_ACCU_DUMMY_IDS (arg_info) = IDS_NEXT (INFO_ACCU_DUMMY_IDS (arg_info));
    
    // Create foldop_fundef based on wl_body
    foldop_fundef = LiftBodyToFoldopFun (TRAVtmpVarName ("CUDA_Fold"), 
                                         wl_body, v1_avis, v2_avis, gpu_res_avis, INFO_GPU_FUNS (arg_info), arg_info);
    INFO_GPU_FUNS (arg_info) = foldop_fundef;
    
    DBUG_PRINT ("  Updating fold partial args");
    foldop_args = CalculateFoldopPartialArgs (foldop_fundef, arg_info);
    FOLD_ARGS (arg_node) = FREEoptFreeTree (FOLD_ARGS (arg_node));
    FOLD_ARGS (arg_node) = foldop_args;
    FOLD_FUNDEF (arg_node) = foldop_fundef;
    
    DBUG_PRINT ("  Updating fold body");
    // Create sdata avis
    INFO_SDATA_ASSIGNS (arg_info) 
        = CreateSdataAssign (neutr_avis, arg_node, INFO_SDATA_ASSIGNS (arg_info), val_size, let_lhs_avis, arg_info);
    sdata_avis = GetFirstIdsAvis (INFO_SDATA_ASSIGNS (arg_info));
    
    // Create sdata assignment
    wl_body = CreateSdataModarrayAssign (sdata_avis, tx_offset_avis, val_avis, wl_body, arg_info);
    sdata_avis = GetLastIdsAvis (wl_body);
    
    // Remove current accu_sel from with loop body
    wl_body = RemoveAssign (wl_body, AVIS_SSAASSIGN (cexpr_avis));
    
    // Create cuda_fold_blocks assign
    INFO_CUDA_FOLD_ASSIGNS (arg_info)
        = CreateFoldPrfAssign (F_cuda_fold_block,
                               TCcreateExprsChainFromAvises (3, INFO_TX_AVIS (arg_info), sdata_avis, neutr_avis),
                               sdata_avis,
                               foldop_fundef,
                               INFO_CUDA_FOLD_ASSIGNS (arg_info), 
                               arg_info);
    sdata_avis = GetFirstIdsAvis (INFO_CUDA_FOLD_ASSIGNS (arg_info));
    
    // Add a cond_wl_assign, so only tx == 0 writes
    INFO_RES_ASSIGNS (arg_info)
        = MakeAssignLet (TBmakeIds (cexpr_avis, NULL),
                         TCmakePrf5 (F_cond_wl_assign,
                                     TBmakeId (INFO_COND_AVIS (arg_info)),
                                     TBmakeNum (0),
                                     TBmakeId (sdata_avis),
                                     TBmakeId (INFO_BX_AVIS (arg_info)),
                                     TBmakeId (fb_avis)),
                         INFO_RES_ASSIGNS (arg_info));
    
    // Update wl body
    CODE_CBLOCK_ASSIGNS (WITH_CODE (INFO_WITH (arg_info))) = wl_body;
    
    DBUG_PRINT ("  Updating the LHS of the first fold");
    // Set LHS to foldbuf
    fb_avis = CopyAvisAndVardec (fb_avis, arg_info);
    IDS_AVIS (INFO_LET_LHS (arg_info)) = fb_avis;
    AVIS_SSAASSIGN (fb_avis) = INFO_ASSIGN (arg_info);
    
    DBUG_PRINT ("  Updating final fold body");
    // Create neutral singleton array
    neutr_arr_expr = TCmakeVector (TYcopyType (cexpr_type),
                                   TBmakeExprs (TBmakeId (neutr_avis), NULL));
    INFO_PREWITH_ASSINGS (arg_info)
        = AppendLetAssign ("neutr_arr",
                           NTCnewTypeCheck_Expr (neutr_arr_expr),
                           neutr_arr_expr,
                           INFO_PREWITH_ASSINGS (arg_info),
                           arg_info);
    
    // Create new fuldbuffer initialised to neutral element
    // Use TBmakeStr to ensure this foldbuf is not alliased
    device_type = CreateDeviceType (cexpr_type);
    INFO_PREWITH_ASSINGS (arg_info)
        = AppendLetAssign ("foldbuf",
                           CreateFoldbufferType (device_type, 1),
                           TCmakePrf2 (F_host2device, 
                                       TBmakeId (GetLastIdsAvis (INFO_PREWITH_ASSINGS (arg_info))),
                                       TBmakeStr (STRcpy (ID_NAME (FOLD_PARTIALMEM (arg_node))))),
                           INFO_PREWITH_ASSINGS (arg_info),
                           arg_info);
    device_type = TYfreeType (device_type);
    fb2_avis = GetLastIdsAvis (INFO_PREWITH_ASSINGS (arg_info));
    
    new_fold = DUPdoDupNode (arg_node);
    FOLD_PARTIALMEM (new_fold) = FREEdoFreeTree (FOLD_PARTIALMEM (new_fold));
    FOLD_PARTIALMEM (new_fold) = TBmakeId (fb2_avis);
    
    // Create new value expression
    wl2_body = AppendLetAssign ("val", 
                                TYcopyType (cexpr_type), 
                                TCmakePrf2 (F_idx_sel, TBmakeId (i_offset_avis), TBmakeId (fb_avis)),
                                wl2_body, 
                                arg_info);
    val2_avis = GetLastIdsAvis (ASSIGN_NEXT (wl2_body));
    
    // Create sdata2 avis
    INFO_FINAL_SDATA_ASSIGNS (arg_info) 
        = CreateSdataAssign (neutr_avis, arg_node, INFO_FINAL_SDATA_ASSIGNS (arg_info), val_size, let_lhs_avis, arg_info);
    sdata2_avis = GetFirstIdsAvis (INFO_FINAL_SDATA_ASSIGNS (arg_info));
    
    // Create sdata assignment
    wl2_body = CreateSdataModarrayAssign (sdata2_avis, i_offset_avis, val2_avis, wl2_body, arg_info);
    sdata2_avis = GetLastIdsAvis (wl2_body);
    // Create cuda_prefold_linear assign
    INFO_FINAL_PREFOLD_ASSIGNS (arg_info)
        = CreateFoldPrfAssign (F_cuda_prefold_linear,
                               TCcreateExprsChainFromAvises (6, 
                                   INFO_FINAL_INDEX_AVIS (arg_info), 
                                   INFO_FINAL_OFFSET_START_AVIS (arg_info), 
                                   INFO_GRID_AVIS (arg_info), 
                                   fb_avis, 
                                   sdata2_avis,
                                   neutr_avis),
                               sdata2_avis,
                               foldop_fundef,
                               INFO_FINAL_PREFOLD_ASSIGNS (arg_info), 
                               arg_info);
    sdata2_avis = GetFirstIdsAvis (INFO_FINAL_PREFOLD_ASSIGNS (arg_info));
    
    // Create cuda_fold_blocks assign
    INFO_FINAL_CUDA_FOLD_ASSIGNS (arg_info)
        = CreateFoldPrfAssign (F_cuda_fold_block,
                               TCcreateExprsChainFromAvises (3, INFO_FINAL_INDEX_AVIS (arg_info), sdata2_avis, neutr_avis),
                               sdata2_avis,
                               foldop_fundef,
                               INFO_FINAL_CUDA_FOLD_ASSIGNS (arg_info), 
                               arg_info);
    sdata2_avis = GetFirstIdsAvis (INFO_FINAL_CUDA_FOLD_ASSIGNS (arg_info));
    
    // Create new cexpr
    INFO_FINAL_RES_ASSIGNS (arg_info)
        = PreppendLetAssign ("res", 
                             TYcopyType (cexpr_type), 
                             TCmakePrf5 (F_cond_wl_assign,
                                         TBmakeId (INFO_FINAL_COND_AVIS (arg_info)),
                                         TBmakeNum (0),
                                         TBmakeId (sdata2_avis),
                                         TBmakeNum (0),
                                         TBmakeId (fb2_avis)), 
                             INFO_FINAL_RES_ASSIGNS (arg_info),
                             arg_info);
    
    cexpr2_avis = GetFirstIdsAvis (INFO_FINAL_RES_ASSIGNS (arg_info));
    WITH_CEXPRS (ASSIGN_RHS (INFO_FINAL_WITH_ASSIGN (arg_info)))
        = TCappendExprs (WITH_CEXPRS (ASSIGN_RHS (INFO_FINAL_WITH_ASSIGN (arg_info))),
                         TBmakeExprs (TBmakeId (cexpr2_avis), NULL));
    
    DBUG_PRINT ("  Updating the LHS of the final fold");
    // Set tmp as lhs of second with
    fb2_avis = CopyAvisAndVardec (fb2_avis, arg_info);
    ASSIGN_LHS (INFO_FINAL_WITH_ASSIGN (arg_info))
        = TCappendIds (ASSIGN_LHS (INFO_FINAL_WITH_ASSIGN (arg_info)), 
                       TBmakeIds (fb2_avis, NULL));
    AVIS_SSAASSIGN (fb2_avis) = INFO_FINAL_WITH_ASSIGN (arg_info);
    
    DBUG_PRINT ("  Appending device2host transfer");
    // Append post assign device2host
    INFO_POSTWITH_ASSINGS (arg_info)
        = AppendLetAssign ("host_foldbuf", 
                           CUconvertDeviceToHostType (AVIS_TYPE (fb2_avis)),
                           TCmakePrf1 (F_device2host, TBmakeId (fb2_avis)),
                           INFO_POSTWITH_ASSINGS (arg_info),
                           arg_info);
    fb2_avis = GetLastIdsAvis (INFO_POSTWITH_ASSINGS (arg_info));
    
    DBUG_PRINT ("  Appending final idx_sel");
    // Append post assign idx_sel
    INFO_POSTWITH_ASSINGS (arg_info) 
        = TCappendAssign (INFO_POSTWITH_ASSINGS (arg_info),
                          MakeAssignLet (TBmakeIds (let_lhs_avis, NULL),
                                         TCmakePrf2 (F_idx_sel,
                                                     TBmakeNum (0),
                                                     TBmakeId (fb2_avis)),
                                         NULL));
    
    // Update wl2 body
    CODE_CBLOCK_ASSIGNS (WITH_CODE (INFO_FINAL_WITH (arg_info))) = wl2_body;
    
    DBUG_PRINT ("  Cleaning up current iteration");
    // Cleanup
    INFO_FUN_VARDECS (arg_info) = NULL;
    INFO_FUN_ARGS (arg_info) = NULL;
    INFO_CEXPRS (arg_info) = EXPRS_NEXT (INFO_CEXPRS (arg_info));
    INFO_LET_LHS (arg_info) = IDS_NEXT (INFO_LET_LHS (arg_info));
    INFO_LUT (arg_info) = LUTremoveContentLut (INFO_LUT (arg_info));
    INFO_LUT_INV (arg_info) = LUTremoveContentLut (INFO_LUT_INV (arg_info));
    
    DBUG_PRINT ("Finished standard fold traversal");
    
    // If final fold, add sdata and cuda_fold
    if (FOLD_NEXT (arg_node) == NULL) {
        // Remove LUTs
        INFO_LUT (arg_info) = LUTremoveLut (INFO_LUT (arg_info));
        INFO_LUT_INV (arg_info) = LUTremoveLut (INFO_LUT_INV (arg_info));    
        
        DBUG_PRINT ("Updating final fold for with-loop"); 
        DBUG_PRINT ("  Adding sdata assign");
        wl_body = TCappendAssign (INFO_SDATA_ASSIGNS (arg_info), wl_body);
        DBUG_PRINT ("  Adding cuda fold assign");
        wl_body = TCappendAssign (wl_body, INFO_CUDA_FOLD_ASSIGNS (arg_info));
        DBUG_PRINT ("  Adding result assigns");
        wl_body = TCappendAssign (wl_body, INFO_RES_ASSIGNS (arg_info));
        CODE_CBLOCK_ASSIGNS (WITH_CODE (INFO_WITH (arg_info))) = wl_body;
        
        DBUG_PRINT ("  Adding final sdata assign");
        wl2_body = TCappendAssign (INFO_FINAL_SDATA_ASSIGNS (arg_info), wl2_body);
        DBUG_PRINT ("  Adding final prefold assign");
        wl2_body = TCappendAssign (wl2_body, INFO_FINAL_PREFOLD_ASSIGNS (arg_info));
        DBUG_PRINT ("  Adding final cuda fold assign");
        wl2_body = TCappendAssign (wl2_body, INFO_FINAL_CUDA_FOLD_ASSIGNS (arg_info));
        DBUG_PRINT ("  Adding final result assigns");
        wl2_body = TCappendAssign (wl2_body, INFO_FINAL_RES_ASSIGNS (arg_info));
        CODE_CBLOCK_ASSIGNS (WITH_CODE (INFO_FINAL_WITH (arg_info))) = wl2_body;
        
        DBUG_PRINT ("Finished final fold for with-loop");
    } else {
        FOLD_NEXT (arg_node) = TRAVdo (FOLD_NEXT (arg_node), arg_info);
    }
    
    // Finally, add new fold
    // Do after next traversal as we have no append fold function
    FOLD_NEXT (new_fold) = WITH_WITHOP (ASSIGN_RHS (INFO_FINAL_WITH_ASSIGN (arg_info)));
    WITH_WITHOP (ASSIGN_RHS (INFO_FINAL_WITH_ASSIGN (arg_info))) = new_fold;    
    
    DBUG_RETURN (arg_node);
}

node *
CUFDcode (node *arg_node, info *arg_info)
{
    node *cexprs, *cexpr_avis, *cexpr_rhs;
    node *accu_expr;
    DBUG_ENTER ();
    
    if (INFO_WITH (arg_info) != NULL) {
        DBUG_PRINT ("Starting code traversal");
        cexprs = CODE_CEXPRS (arg_node);
        while (cexprs != NULL) {
            cexpr_avis = ID_AVIS (EXPRS_EXPR1 (cexprs));
            cexpr_rhs = ASSIGN_RHS (AVIS_SSAASSIGN (cexpr_avis));
            
            DBUG_ASSERT (NODE_TYPE (cexpr_rhs) == N_prf && 
                         PRF_PRF (cexpr_rhs) == F_accu_sel,
                         "Cexprs is not defined by accu_sel.");
            
            // Set cexpr expr to regular accu fold result
            DBUG_PRINT ("  Changing cexpr expression");
            accu_expr = EXPRS_EXPR1 (PRF_ARGS (cexpr_rhs));
            EXPRS_EXPR1 (cexprs) = FREEdoFreeTree (EXPRS_EXPR1 (cexprs));
            EXPRS_EXPR1 (cexprs) = DUPdoDupTree (accu_expr);
            cexprs = EXPRS_NEXT (cexprs);
        }
        DBUG_PRINT ("Finished code traversal");
        
        CODE_NEXT (arg_node) = TRAVopt (CODE_NEXT (arg_node), arg_info);
    } else {
        CODE_CBLOCK (arg_node) = TRAVopt (CODE_CBLOCK (arg_node), arg_info);
        CODE_NEXT (arg_node) = TRAVopt (CODE_NEXT (arg_node), arg_info);
    }
    
    DBUG_RETURN (arg_node);
}
/** <!--********************************************************************-->
 * @}  <!-- Traversal functions -->
 *****************************************************************************/


#undef DBUG_PREFIX
